//+------------------------------------------------------------------+
//|                                                  ExpertTools.mqh |
//|                                  Copyright 2023, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2023, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#include <Trade\Trade.mqh>
#include <Trade\SymbolInfo.mqh>
#include <Trade\AccountInfo.mqh>
#include <ErrorDescription.mqh>
//+------------------------------------------------------------------+
//|Open trade                                                        |
//+------------------------------------------------------------------+
bool SendOrder(string sy,double price,ENUM_ORDER_TYPE type,double tradevolume,ulong slp,double stoploss,double takeprofit,string tradecomment,ulong magic_num,datetime expire=NULL)
  {
//---  
   if(StringLen(sy)<1)
    sy = _Symbol; 
//---  
   CTrade trade;
   CSymbolInfo syminfo;
   CAccountInfo accinfo;
//---
   trade.SetExpertMagicNumber(magic_num); // magic
   trade.SetMarginMode();
   trade.SetDeviationInPoints(slp);
//---   
   if(!trade.SetTypeFillingBySymbol(sy))
     {
      Print(__FUNCTION__," Unknown Type filling mode for ", sy);
      return false;
     }
//---
   if( !syminfo.Name(sy))
     {
      Print(__FUNCTION__," : ",__LINE__," : symbol info config error, ",sy);
      return false;
     } 
//--- check for right lots amount
   if(tradevolume<syminfo.LotsMin() || tradevolume>syminfo.LotsMax())
     {
      Print("Lots amount for "+sy+" order must be in the range from ",syminfo.LotsMin()," to ",syminfo.LotsMax());
      return(false);
     }
   if(MathAbs(tradevolume/syminfo.LotsStep()-MathRound(tradevolume/syminfo.LotsStep()))>1.0E-10)
     {
      Print("Lots amount for "+sy+" order does not correspond with lot step ",syminfo.LotsStep());
      return(false);
     }            
//---
   bool ans=false;
   syminfo.RefreshRates();
//---   
   if(type == ORDER_TYPE_BUY || type == ORDER_TYPE_SELL)
     price = (type == ORDER_TYPE_BUY)?syminfo.Ask():syminfo.Bid();
//---     
   price = NormalizeDouble(MathAbs(price),syminfo.Digits());  
//---
   if(stoploss)
    {
     if(int(MathAbs(price - MathAbs(stoploss))/syminfo.Point()) < syminfo.StopsLevel())
       {
        Print(__FUNCSIG__," : "," Invalid stoploss level ", sy);
        return false; 
       }
    }
//---
   if(takeprofit)
    {
     if(int(MathAbs(price - MathAbs(takeprofit))/syminfo.Point()) < syminfo.StopsLevel())
       {
        Print(__FUNCSIG__," : "," Invalid takeprofit level ", sy);
        return false; 
       }
    }         
//---
   switch(type)
     {
      case ORDER_TYPE_BUY:
         if(accinfo.FreeMarginCheck(sy,type,tradevolume,price)<0.0)
           {
            Print("Insufficient funds to open long order ("+sy+"). Free Margin = ", accinfo.FreeMargin());
            return false;
           } 
         ans=trade.Buy(tradevolume,sy,price,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),tradecomment);
         break;
      case ORDER_TYPE_SELL:
         if(accinfo.FreeMarginCheck(sy,type,tradevolume,price)<0.0)
           {
            Print("Insufficient funds to open short order ("+sy+"). Free Margin = ", accinfo.FreeMargin());
            return false;
           } 
         ans=trade.Sell(tradevolume,sy,price,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),tradecomment);
         break;
      case ORDER_TYPE_SELL_LIMIT:
         if((price - syminfo.Bid())/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid sell limit price level ", sy);
           return false;
          }
         ans=trade.SellLimit(tradevolume,price,sy,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
      case ORDER_TYPE_BUY_LIMIT:
         if((syminfo.Ask() - price)/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid buy limit price level ", sy);
           return false;
          }
         ans=trade.BuyLimit(tradevolume,price,sy,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
      case ORDER_TYPE_SELL_STOP:
         if((syminfo.Bid() - price)/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid sell stop price level ", sy);
           return false;
          }
         ans=trade.SellStop(tradevolume,price,sy,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
      case ORDER_TYPE_BUY_STOP:
         if((price - syminfo.Ask())/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid buy stop price level ", sy);
           return false;
          }
         ans=trade.BuyStop(tradevolume,price,sy,NormalizeDouble(stoploss,syminfo.Digits()),NormalizeDouble(takeprofit,syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
     }  
//--- 
   trade.PrintResult();
//---       
   return(ans);
  }
//+------------------------------------------------------------------+
//|Open trade                                                        |
//+------------------------------------------------------------------+
bool SendRiskCalOrder(string sy,double price,ENUM_ORDER_TYPE type,double riskpercent,ulong slp,ulong stoplosspoints,ulong takeprofitpoints,string tradecomment,ulong magic_num,datetime expire=NULL)
  {
//---  
   if(StringLen(sy)<1)
    sy = _Symbol; 
//---  
   CTrade trade;
   CSymbolInfo syminfo;
   CAccountInfo accinfo;
   double tradevolume,stoploss=0,takeprofit=0;
//---
   trade.SetExpertMagicNumber(magic_num); // magic
   trade.SetMarginMode();
   trade.SetDeviationInPoints(slp);
//---   
   if(!trade.SetTypeFillingBySymbol(sy))
     {
      Print(__FUNCTION__," Unknown Type filling mode for ", sy);
      return false;
     }
//---
   if( !syminfo.Name(sy))
     {
      Print(__FUNCTION__," : ",__LINE__," : symbol info config error, ",sy);
      return false;
     } 
//---
   bool ans=false;
   syminfo.RefreshRates();
//---   
   if(type == ORDER_TYPE_BUY || type == ORDER_TYPE_SELL)
     price = (type == ORDER_TYPE_BUY)?syminfo.Ask():syminfo.Bid();
//---     
   price = NormalizeDouble(MathAbs(price),syminfo.Digits()); 
//---     
   if(price<=0)
    {
     Print(__FUNCTION__," Invalid price specified ");
     return false;
    } 
//---     
   if(riskpercent<=0)
    {
     Print(__FUNCTION__," Invalid risk percent specified");
     return false;
    }           
//---
   stoploss = (stoplosspoints)?double(stoplosspoints)*syminfo.Point():0;
   takeprofit = (takeprofitpoints)?double(takeprofitpoints)*syminfo.Point():0;
//---
   if(stoplosspoints)
    {
     if(int(stoplosspoints) < syminfo.StopsLevel())
       {
        Print(__FUNCSIG__," : "," Invalid stoploss level ", sy);
        return false; 
       }
       
     tradevolume = CalculatePositionSize(stoploss,riskpercent,syminfo);  
      
    }
   else
    {
     Print("Invalid input for stoplosspoints. Cannot be 0");
     return false;
    } 
//---
   if(takeprofitpoints)
    {
     if(int(takeprofitpoints) < syminfo.StopsLevel())
       {
        Print(__FUNCSIG__," : "," Invalid takeprofit level ", sy);
        return false; 
       }
    }         
//---      

//---
   switch(type)
     {
      case ORDER_TYPE_BUY:
          stoploss = (stoplosspoints)?price - stoploss:0;
          takeprofit = (takeprofitpoints)?price + takeprofit:0;
         if(accinfo.FreeMarginCheck(sy,type,tradevolume,price)<0.0)
           {
            Print("Insufficient funds to open long order ("+sy+"). Free Margin = ", accinfo.FreeMargin());
            return false;
           } 
         ans=trade.Buy(tradevolume,sy,price,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),tradecomment);
         break;
      case ORDER_TYPE_SELL:
          stoploss = (stoplosspoints)?price + stoploss:0;
          takeprofit = (takeprofitpoints)?price - takeprofit:0;
         if(accinfo.FreeMarginCheck(sy,type,tradevolume,price)<0.0)
           {
            Print("Insufficient funds to open short order ("+sy+"). Free Margin = ", accinfo.FreeMargin());
            return false;
           } 
         ans=trade.Sell(tradevolume,sy,price,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),tradecomment);
         break;
      case ORDER_TYPE_SELL_LIMIT:
          stoploss = (stoplosspoints)?price + stoploss:0;
          takeprofit = (takeprofitpoints)?price - takeprofit:0;
         if((price - syminfo.Bid())/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid sell limit price level ", sy);
           return false;
          }
         ans=trade.SellLimit(tradevolume,price,sy,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
      case ORDER_TYPE_BUY_LIMIT:
          stoploss = (stoplosspoints)?price - stoploss:0;
          takeprofit = (takeprofitpoints)?price + takeprofit:0;
         if((syminfo.Ask() - price)/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid buy limit price level ", sy);
           return false;
          }
         ans=trade.BuyLimit(tradevolume,price,sy,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
      case ORDER_TYPE_SELL_STOP:
          stoploss = (stoplosspoints)?price + stoploss:0;
          takeprofit = (takeprofitpoints)?price - takeprofit:0;
         if((syminfo.Bid() - price)/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid sell stop price level ", sy);
           return false;
          }
         ans=trade.SellStop(tradevolume,price,sy,NormalizeDouble(MathAbs(stoploss),syminfo.Digits()),NormalizeDouble(MathAbs(takeprofit),syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
      case ORDER_TYPE_BUY_STOP:
          stoploss = (stoplosspoints)?price - stoploss:0;
          takeprofit = (takeprofitpoints)?price + takeprofit:0;
         if((price - syminfo.Ask())/syminfo.Point() < syminfo.StopsLevel())
          {
           Print("Invalid buy stop price level ", sy);
           return false;
          }
         ans=trade.BuyStop(tradevolume,price,sy,NormalizeDouble(stoploss,syminfo.Digits()),NormalizeDouble(takeprofit,syminfo.Digits()),(expire!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,expire,tradecomment);
         break;
     }  
//--- 
   trade.PrintResult();
//---       
   return(ans);
  }  
//+------------------------------------------------------------------+
//| Close market order                                               |
//+------------------------------------------------------------------+
bool CloseOrder(ulong magic_num,string sy=NULL,ulong positionticket=0,ulong slp=10,double tradevolume=0)
  {
//---  
   CTrade trade;
   CSymbolInfo syminfo;
//---   
   trade.SetExpertMagicNumber(magic_num);
//---
   if(sy==NULL && !positionticket)
    {
     Print(" Ambiguous call. Specify either symbol or the position ticket");
     return false;
    }
//---   
   bool  result = false;;
//---
   if(tradevolume<=0)
      result = (positionticket)?trade.PositionClose(positionticket,slp):(sy!=NULL)?trade.PositionClose(sy,slp):false;
   else
     { 
      if(sy!=NULL && syminfo.Name(sy))
       {
        if(MathAbs(tradevolume/syminfo.LotsStep()-MathRound(tradevolume/syminfo.LotsStep()))>1.0E-10)
          {
           Print("Lots amount for "+sy+" order does not correspond with lot step ",syminfo.LotsStep());
           return(false);
          }     
       }
      result = (positionticket)?trade.PositionClosePartial(positionticket,tradevolume,slp):(sy!=NULL)?trade.PositionClosePartial(sy,tradevolume,slp):false;
     }
//---
   trade.PrintResult();
//--- 
   return result;   
  } 
//+------------------------------------------------------------------+
//|Modif Order                                                       |
//+------------------------------------------------------------------+
bool ModifyOrder(ulong magic_num,string sy,ulong ticket,double stoploss,double takeprofit,double new_pending_price_level=0,datetime new_expiration=NULL)
  {
//---  
   CTrade trade;
   CSymbolInfo syminfo;
//---   
   trade.SetExpertMagicNumber(magic_num);
   bool result = false;
//---
   if(syminfo.Name(sy))
    {
     stoploss = NormalizeDouble(MathAbs(stoploss),syminfo.Digits());
     takeprofit = NormalizeDouble(MathAbs(takeprofit),syminfo.Digits());
     new_pending_price_level = NormalizeDouble(MathAbs(new_pending_price_level),syminfo.Digits());
    }   
//---
   if(new_pending_price_level>0 && ticket)
     {
      if(syminfo.Name()!=NULL)
       {
        if((takeprofit  && MathAbs(new_pending_price_level - takeprofit)/syminfo.Point() < syminfo.StopsLevel()) ||
           (stoploss    && MathAbs(new_pending_price_level - stoploss)/syminfo.Point() < syminfo.StopsLevel()) ) 
         {
          Print(__FUNCTION__, " Invalid stops ", sy);
          return false;
         }    
       }
      result = trade.OrderModify(ticket,new_pending_price_level,stoploss,takeprofit,(new_expiration!=NULL)?ORDER_TIME_SPECIFIED:ORDER_TIME_GTC,new_expiration);
     }   
   else
      result = (ticket)?trade.PositionModify(ticket,stoploss,takeprofit):(sy!=NULL)?trade.PositionModify(sy,stoploss,takeprofit):false; 
//---
   trade.PrintResult();
//---
   return result;     
//---
  }
//+------------------------------------------------------------------+
//|calculate optimal position size                                   |
//+------------------------------------------------------------------+
double CalculatePositionSize(const double Stop,const double risk, CSymbolInfo &syminfo)
{
//---    
 double RiskMoney=0.0, PositionSize=0.0,bb=AccountInfoDouble(ACCOUNT_BALANCE);
//---     
 RiskMoney = RoundDown(bb * risk / 100, 2);
//--- 
 double UnitCost = syminfo.TickValue();
//---  
 if ((Stop != 0) && (UnitCost != 0)) 
     PositionSize = RoundDown(RiskMoney / (Stop * UnitCost / syminfo.TickSize()), 2);
 else
   {
    Print(__FUNCTION__," Error calculating optimal lotsize");
    return EMPTY_VALUE;
   }
//---      
 if (PositionSize < syminfo.LotsMin()) 
    PositionSize = syminfo.LotsMin();
 else 
   if (PositionSize > syminfo.LotsMax()) 
       PositionSize = syminfo.LotsMax();
//---       
 double steps = (syminfo.LotsStep()>0)?PositionSize / syminfo.LotsStep():0;
//---
 if (MathFloor(steps) < steps) 
    PositionSize = MathFloor(steps) * syminfo.LotsStep();
//--- 
 return(PositionSize);
}
//+------------------------------------------------------------------+
//| Round down function                                              |
//+------------------------------------------------------------------+
double RoundDown(const double value, const double digits)
{
   int norm = (int)MathPow(10, digits);
   return(MathFloor(value * norm) / norm);
}
//+------------------------------------------------------------------+
//|Close all market orders                                           |
//+------------------------------------------------------------------+
uint CloseAll(ulong magic_number=ULONG_MAX,const string sym=NULL,const ENUM_POSITION_TYPE ordertype=WRONG_VALUE,const ulong mslip=10) 
{
 CPositionInfo posinfo;
//---
 uint countclosed=0;
//--- 
 for(int i = PositionsTotal()-1; i>-1;i--)
  {
   if(!posinfo.SelectByIndex(i))
     continue;
   if(magic_number<ULONG_MAX && posinfo.Magic()!=long(magic_number))
     continue;
   if(sym!=NULL && posinfo.Symbol()!=sym)
     continue;
   if(ordertype!=WRONG_VALUE && posinfo.PositionType()!=ordertype)
     continue;
   if(CloseOrder(ulong(posinfo.Magic()),posinfo.Symbol(),posinfo.Ticket()))
      countclosed++;
  }
//---
 return countclosed;
} 
//+------------------------------------------------------------------+
//| Delete all pending orders                                        |
//+------------------------------------------------------------------+
uint DeleteAll(ulong magic_number=ULONG_MAX,const string sym=NULL,const ENUM_ORDER_TYPE ordertype=WRONG_VALUE)
{
 COrderInfo orderinfo;
//--- 
 uint countdeleted=0;
//--- declare and initialize the trade request and result of trade request
   MqlTradeRequest request={};
   MqlTradeResult  result={}; 
//---
 for(int i = OrdersTotal()-1; i>-1;i--)
  {
   if(!orderinfo.SelectByIndex(i))
     continue;
   if(orderinfo.OrderType()<2)
     continue;  
   if(magic_number<ULONG_MAX && orderinfo.Magic()!=long(magic_number))
     continue;
   if(ordertype!=WRONG_VALUE && orderinfo.OrderType()!=ordertype)
     continue;
   if(sym!=NULL && orderinfo.Symbol()!=sym)
     continue;
//--- zeroing the request and result values
   ZeroMemory(request);
   ZeroMemory(result);
//--- setting the operation parameters     
   request.action=TRADE_ACTION_REMOVE;                   // type of trade operation
   request.order = orderinfo.Ticket();        
   if(OrderSend(request,result))
     countdeleted++;
   else
     Print("Failed to delete order ",request.order, " Error ", ErrorDescription(result.retcode));  
  } 
//---
  return countdeleted;
}
//+------------------------------------------------------------------+
//|time of last market or pending entry                              |
//+------------------------------------------------------------------+
datetime LastOrderOpenTime(long &lastpositionID,const string sym = NULL, const ulong magic_number = ULONG_MAX)
{
 CDealInfo dealinfo;
//---
 datetime transactiontime=0;
//--- 
 HistorySelect(0,TimeCurrent());
//--- 
 for(int i = HistoryDealsTotal()-1;i>-1;i--)
  {
   if(!dealinfo.SelectByIndex(i))
     continue;
   if(dealinfo.Entry()!=DEAL_ENTRY_IN)
     continue;
   if(magic_number<ULONG_MAX && dealinfo.Magic()!=long(magic_number))
     continue;
   if(sym!=NULL && dealinfo.Symbol()!=sym)
     continue;
   transactiontime = dealinfo.Time();
   lastpositionID  = dealinfo.PositionId();
   break;        
  }
//---  
 COrderInfo orderinfo;
//---
 for(int i = OrdersTotal()-1; i>-1;i--)
  {
   if(!orderinfo.SelectByIndex(i))
     continue;
   if(orderinfo.OrderType()<2)
     continue;  
   if(magic_number<ULONG_MAX && orderinfo.Magic()!=long(magic_number))
     continue;
   if(sym!=NULL && orderinfo.Symbol()!=sym)
     continue;
   transactiontime = (orderinfo.TimeSetup()>transactiontime)?orderinfo.TimeSetup():transactiontime;
   break; 
  } 
//---
 return transactiontime;
}
//+------------------------------------------------------------------+
//|Enumerate all market orders                                       |
//+------------------------------------------------------------------+
uint SumMarketOrders(ulong magic_number=ULONG_MAX,const string sym=NULL,const ENUM_POSITION_TYPE ordertype=WRONG_VALUE) 
{
 CPositionInfo posinfo;
//---
 uint count=0;
//--- 
 for(int i = PositionsTotal()-1; i>-1;i--)
  {
   if(!posinfo.SelectByIndex(i))
     continue;
   if(magic_number<ULONG_MAX && posinfo.Magic()!=long(magic_number))
     continue;
   if(sym!=NULL && posinfo.Symbol()!=sym)
     continue;
   if(ordertype!=WRONG_VALUE && posinfo.PositionType()!=ordertype)
     continue;
   count++;
  }
//---
 return count;
} 